/* Nefarious IRCu functions
 *
 * (C) 2005 Neil Spierling <sirvulcan@gmail.com>
 *
 * Please read COPYING and README for further details.
 *
 * Based on the original code of Epona by Lara.
 * Based on the original code of Services by Andy Church. 
 * 
 *
 */

#include "services.h"
#include "pseudo.h"
#include "nefarious.h"

int p10nickcnt = 0;
int p10servcnt = 1;
int p10inuse = 0;

IRCDVar myIrcd[] = {
    {"Nefarious IRCu 0.5.0+",   /* ircd name */
     "+ok",                     /* nickserv mode */
     "+ok",                     /* chanserv mode */
     "+ok",                     /* memoserv mode */
     "ok",                      /* hostserv mode */
     "+iok",                    /* operserv mode */
     "+ok",                     /* botserv mode  */
     "+k",                      /* helpserv mode */
     "+ik",                     /* Dev/Null mode */
     "+iok",                    /* Global mode   */
     "+ok",                     /* nickserv alias mode */
     "+ok",                     /* chanserv alias mode */
     "+ok",                     /* memoserv alias mode */
     "+ok",                     /* hostserv alias mode */
     "+io",                     /* operserv alias mode */
     "+ok",                     /* botserv alias mode  */
     "+k",                      /* helpserv alias mode */
     "+ik",                     /* Dev/Null alias mode */
     "+iok",                    /* Global alias mode   */
     "+k",                      /* Used by BotServ Bots */
     3,                         /* Chan Max Symbols     */
     "-ilmnprst",               /* Modes to Remove */
     "+o",                      /* Channel Umode used by Botserv bots */
     1,                         /* SVSNICK */
     1,                         /* Vhost  */
     0,                         /* Has Owner */
     NULL,                      /* Mode to set for an owner */
     NULL,                      /* Mode to unset for an owner */
     NULL,                      /* Mode to set for admin */
     NULL,                      /* Mode to unset for admin */
     NULL,                      /* Mode On Reg          */
     NULL,                      /* Mode on UnReg        */
     NULL,                      /* Mode on Nick Change  */
     0,                         /* Supports SGlines     */
     0,                         /* Supports SQlines     */
     0,                         /* Supports SZlines     */
     1,                         /* Supports Halfop +h   */
     3,                         /* Number of server args */
     0,                         /* Join 2 Set           */
     0,                         /* Join 2 Message       */
     1,                         /* Has exceptions +e    */
     0,                         /* TS Topic Forward     */
     0,                         /* TS Topci Backward    */
     0,                         /* Protected Umode      */
     0,                         /* Has Admin            */
     0,                         /* Chan SQlines         */
     0,                         /* Quit on Kill         */
     0,                         /* SVSMODE unban        */
     0,                         /* Has Protect          */
     0,                         /* Reverse              */
     0,                         /* Chan Reg             */
     0,                         /* Channel Mode         */
     1,                         /* vidents              */
     0,                         /* svshold              */
     0,                         /* time stamp on mode   */
     1,                         /* NICKIP               */
     0,                         /* UMODE                */
     0,                         /* O:LINE               */
     0,                         /* VHOST ON NICK        */
     0,                         /* Change RealName      */
     0,                         /* No Knock             */
     CMODE_a,                   /* Admin Only           */
     DEFAULT_MLOCK,             /* Default MLOCK        */
     UMODE_x,                   /* Vhost Mode           */
     0,                         /* +f                   */
     0,                         /* +L                   */
     0,                         /* +f Mode                          */
     0,                         /* +L Mode                              */
     0,                         /* On nick change check if they could be identified */
     0,                         /* No Knock requires +i */
     NULL,                      /* CAPAB Chan Modes             */
     0,                         /* We support TOKENS */
     0,                         /* TOKENS are CASE inSensitive */
     0,                         /* TIME STAMPS are BASE64 */
     0,                         /* +I support */
     0,                         /* SJOIN ban char */
     0,                         /* SJOIN except char */
     0,                         /* SJOIN invite char */
     0,                         /* Can remove User Channel Modes with SVSMODE */
     0,                         /* Sglines are not enforced until user reconnects */
     "x",                      /* vhost char */
     0,                         /* ts6 */
     0,                         /* support helper umode */
     1,                         /* p10 */
     NULL,                      /* character set */
     1,                         /* reports sync state */
     }
    ,
    {NULL}
};

IRCDCAPAB myIrcdcap[] = {
    {
     1,                         /* NOQUIT       */
     0,                         /* TSMODE       */
     0,                         /* UNCONNECT    */
     0,                         /* NICKIP       */
     0,                         /* SJOIN        */
     0,                         /* ZIP          */
     0,                         /* BURST        */
     0,                         /* TS5          */
     0,                         /* TS3          */
     0,                         /* DKEY         */
     0,                         /* PT4          */
     0,                         /* SCS          */
     0,                         /* QS           */
     0,                         /* UID          */
     0,                         /* KNOCK        */
     0,                         /* CLIENT       */
     0,                         /* IPV6         */
     0,                         /* SSJ5         */
     0,                         /* SN2          */
     0,                         /* TOKEN        */
     0,                         /* VHOST        */
     0,                         /* SSJ3         */
     0,                         /* NICK2        */
     0,                         /* UMODE2       */
     0,                         /* VL           */
     0,                         /* TLKEXT       */
     0,                         /* DODKEY       */
     0,                         /* DOZIP        */
     0, 0, 0}
};


/* ABAAA M +i */
void nefarious_set_umode(User * user, int ac, char **av)
{
    int add = 1;                /* 1 if adding modes, 0 if deleting */
    int h = 1;
    char *modes = av[0];
    char *sethost = av[1];
    char *uhb, *uh = NULL;
    char *vhost = NULL;
    char *vident = NULL;
    User *sender;

    sender = user;

    ac--;

    if (debug)
        alog("debug: Changing mode for %s to %s", sender->nick, modes);

    while (*modes) {

        /* This looks better, much better than "add ? (do_add) : (do_remove)".
         * At least this is readable without paying much attention :) -GD
         */
        if (add)
            sender->mode |= umodes[(int) *modes];
        else
            sender->mode &= ~umodes[(int) *modes];

        switch (*modes++) {
        case '+':
            add = 1;
            break;
        case '-':
            add = 0;
            break;
        case 'h':
            if (add) {
                for (uh = strtok_r(sethost, "@", &uhb);
                    uh;
                    uh = strtok_r(NULL, "@", &uhb)) {
                   if (h == 1)
                       vident = uh;
                   else if (h == 2)
                       vhost = uh;
                   h++;
                }
                h = 1;

                if (vident)
                    change_user_username(sender, vident);

                change_user_host(sender, vhost ? vhost : av[1]);
            } else {
                free(sender->vhost);
            }
            break;
        case 'o':
            if (add) {
                opcnt++;

                if (WallOper)
                    anope_cmd_global(s_OperServ,
                                     "\2%s\2 is now an IRC operator.",
                                     sender->nick);
                display_news(sender, NEWS_OPER);

            } else {
                opcnt--;
            }
            break;
        case 'r':
            if (add && !nick_identified(user)) {
                send_cmd(p10id, "AC %s U", user->uid);
                user->mode &= ~UMODE_r;
            }
        }

    }
}

unsigned long umodes[128] = {
    0, 0, 0,                    /* Unused */
    0, 0, 0,                    /* Unused */
    0, 0, 0,                    /* Unused, Unused, Horzontal Tab */
    0, 0, 0,                    /* Line Feed, Unused, Unused */
    0, 0, 0,                    /* Carriage Return, Unused, Unused */
    0, 0, 0,                    /* Unused */
    0, 0, 0,                    /* Unused */
    0, 0, 0,                    /* Unused */
    0, 0, 0,                    /* Unused */
    0, 0, 0,                    /* Unused */
    0, 0, 0,                    /* Unused, Unused, Space */
    0, 0, 0,                    /* ! " #  */
    0, 0, 0,                    /* $ % &  */
    0, 0, 0,                    /* ! ( )  */
    0, 0, 0,                    /* * + ,  */
    0, 0, 0,                    /* - . /  */
    0, 0,                       /* 0 1 */
    0, 0,                       /* 2 3 */
    0, 0,                       /* 4 5 */
    0, 0,                       /* 6 7 */
    0, 0,                       /* 8 9 */
    0, 0,                       /* : ; */
    0, 0, 0,                    /* < = > */
    0, 0,                       /* ? @ */
    0, UMODE_B, 0,              /* A B C */
    0, 0, 0,                    /* D E F */
    0, 0, UMODE_I,              /* G H I */
    0, 0, 0,                    /* J K L */
    0, 0, 0,                    /* M N O */
    0, 0, UMODE_R,              /* P Q R */
    0, 0, 0,                    /* S T U */
    0, 0, UMODE_X,              /* V W X */
    0,                          /* Y */
    0,                          /* Z */
    0, 0, 0,                    /* [ \ ] */
    0, 0, 0,                    /* ^ _ ` */
    UMODE_a, 0, 0,              /* a b c */
    UMODE_d, 0, UMODE_f,        /* d e f */
    UMODE_g, UMODE_h, UMODE_i,  /* g h i */
    0, UMODE_k, 0,              /* j k l */
    0, UMODE_n, UMODE_o,        /* m n o */
    0, 0, UMODE_r,              /* p q r */
    UMODE_s, 0, 0,              /* s t u */
    0, UMODE_w, UMODE_x,        /* v w x */
    0,                          /* y */
    0,                          /* z */
    0, 0, 0,                    /* { | } */
    0, 0                        /* ~  */
};


char myCsmodes[128] = {
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    0,
    0,                        /* (33) !  */
    0, 0, 0,
    'h',
    'b',
    0, 0, 0,
    0,

    'v', 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    'o', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,

    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

CMMode myCmmodes[128] = {
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL},     /* BCD */
    {NULL}, {NULL}, {NULL},     /* EFG */
    {NULL}, {NULL}, {NULL},     /* HIJ */
    {NULL}, {NULL}, {NULL},     /* KLM */
    {NULL}, {NULL}, {NULL},     /* NOP */
    {NULL}, {NULL}, {NULL},     /* QRS */
    {NULL}, {NULL}, {NULL},     /* TUV */
    {NULL}, {NULL}, {NULL},     /* WXY */
    {NULL},                     /* Z */
    {NULL}, {NULL},             /* (char 91 - 92) */
    {NULL}, {NULL}, {NULL},     /* (char 93 - 95) */
    {NULL},                     /* ` (char 96) */
    {NULL},                     /* a (char 97) */
    {add_ban, del_ban},
    {NULL},
    {NULL},
    {add_exception, del_exception},
    {NULL},
    {NULL},
    {NULL}, {NULL}, {NULL}, {NULL}, {NULL}, {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL}, {NULL}, {NULL}, {NULL}, {NULL}, {NULL},
    {NULL}, {NULL}, {NULL}, {NULL}, {NULL}, {NULL}, {NULL}, {NULL}
};


CBMode myCbmodes[128] = {
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0},
    {0},                        /* A */
    {0},                        /* B */
    {CMODE_C, 0, NULL, NULL},
    {0},                        /* D */
    {0},                        /* E */
    {0},                        /* F */
    {0},                        /* G */
    {0},                        /* H */
    {0},                        /* I */
    {0},                        /* J */
    {0},                        /* K */
    {CMODE_L, 0, NULL, NULL},
    {CMODE_M, 0, NULL, NULL},
    {CMODE_N, 0, NULL, NULL},
    {CMODE_O, CBM_NO_USER_MLOCK, NULL, NULL},
    {0},
    {CMODE_Q, 0, NULL, NULL},
    {0},
    {CMODE_S, 0, NULL, NULL},
    {CMODE_T, 0, NULL, NULL},
    {0},                        /* U */
    {0},                        /* V */
    {0},                        /* W */
    {0},                        /* X */
    {0},                        /* Y */
    {CMODE_Z, 0, NULL, NULL},
    {0}, {0}, {0}, {0}, {0}, {0},
    {CMODE_a, CBM_NO_USER_MLOCK, NULL, NULL},
    {0},                        /* b */
    {CMODE_c, 0, NULL, NULL},
    {0},                        /* d */
    {0},                        /* e */
    {0},                        /* f */
    {0},                        /* g */
    {0},                        /* h */
    {CMODE_i, 0, NULL, NULL},
    {0},                        /* j */
    {CMODE_k, 0, set_key, cs_set_key},
    {CMODE_l, CBM_MINUS_NO_ARG, set_limit, cs_set_limit},
    {CMODE_m, 0, NULL, NULL},
    {CMODE_n, 0, NULL, NULL},
    {0},
    {CMODE_p, 0, NULL, NULL},
    {0},
    {CMODE_r, 0, NULL, NULL},
    {CMODE_s, 0, NULL, NULL},
    {CMODE_t, 0, NULL, NULL},
    {0},
    {0},                        /* v */
    {0},                        /* w */
    {0},                        /* x */
    {0},                        /* y */
    {CMODE_z, 0, NULL, NULL},
    {0}, {0}, {0}, {0}
};

CBModeInfo myCbmodeinfos[] = {
    {'a', CMODE_a, 0, NULL, NULL},
    {'c', CMODE_c, 0, NULL, NULL},
    {'i', CMODE_i, 0, NULL, NULL},
    {'k', CMODE_k, 0, get_key, cs_get_key},
    {'l', CMODE_l, CBM_MINUS_NO_ARG, get_limit, cs_get_limit},
    {'m', CMODE_m, 0, NULL, NULL},
    {'n', CMODE_n, 0, NULL, NULL},
    {'p', CMODE_p, 0, NULL, NULL},
    {'r', CMODE_r, 0, NULL, NULL},
    {'s', CMODE_s, 0, NULL, NULL},
    {'t', CMODE_t, 0, NULL, NULL},
    {'z', CMODE_z, 0, NULL, NULL},
    {'C', CMODE_C, 0, NULL, NULL},
    {'L', CMODE_L, 0, NULL, NULL},
    {'M', CMODE_M, 0, NULL, NULL},
    {'N', CMODE_N, 0, NULL, NULL},
    {'O', CMODE_O, 0, NULL, NULL},
    {'Q', CMODE_Q, 0, NULL, NULL},
    {'S', CMODE_S, 0, NULL, NULL},
    {'T', CMODE_T, 0, NULL, NULL},
    {'Z', CMODE_Z, 0, NULL, NULL},
    {0}
};


CUMode myCumodes[128] = {
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},

    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},

    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},
    {0}, {0}, {0}, {0}, {0}, {0}, {0}, {0},

    {0},

    {0},                        /* a */
    {0},                        /* b */
    {0},                        /* c */
    {0},                        /* d */
    {0},                        /* e */
    {0},                        /* f */
    {0},                        /* g */
    {CUS_HALFOP, 0, check_valid_op},
    {0},                        /* i */
    {0},                        /* j */
    {0},                        /* k */
    {0},                        /* l */
    {0},                        /* m */
    {0},                        /* n */
    {CUS_OP, CUF_PROTECT_BOTSERV, check_valid_op},
    {0},                        /* p */
    {0},                        /* q */
    {0},                        /* r */
    {0},                        /* s */
    {0},                        /* t */
    {0},                        /* u */
    {CUS_VOICE, 0, NULL},
    {0},                        /* w */
    {0},                        /* x */
    {0},                        /* y */
    {0},                        /* z */
    {0}, {0}, {0}, {0}, {0}
};



void nefarious_cmd_notice(char *source, char *dest, char *buf)
{
    User *u;
    Uid *ud;

    if (!buf) {
        return;
    }

    u = finduser(dest);
    ud = find_uid(source);

    if (UsePrivmsg) {
        nefarious_cmd_privmsg(source, dest, buf);
    } else {
        send_cmd((ud ? ud->uid : source), "O %s :%s", (u ? u->uid : dest),
                 buf);
    }
}

void nefarious_cmd_notice2(char *source, char *dest, char *msg)
{
    Uid *ud;
    User *u;

    ud = find_uid(source);
    u = finduser(dest);
    send_cmd((ud ? ud->uid : source), "O %s :%s", (u ? u->uid : dest),
             msg);
}

void nefarious_cmd_privmsg(char *source, char *dest, char *buf)
{
    User *u;
    Uid *ud;

    if (!buf) {
        return;
    }

    u = finduser(dest);
    ud = find_uid(source);

    send_cmd((ud ? ud->uid : source), "P %s :%s", (u ? u->uid : dest),
             buf);
}

void nefarious_cmd_privmsg2(char *source, char *dest, char *msg)
{
    Uid *ud, *ud2;

    ud = find_uid(source);
    ud2 = find_uid(dest);

    send_cmd((ud ? ud->uid : source), "P %s :%s",
             (ud2 ? ud2->uid : dest), msg);
}

void nefarious_cmd_serv_notice(char *source, char *dest, char *msg)
{
    User *u;
    Uid *ud;

    u = finduser(dest);
    ud = find_uid(source);

    send_cmd((ud ? ud->uid : source), "O $%s :%s", (u ? u->uid : dest), msg);
}

void nefarious_cmd_serv_privmsg(char *source, char *dest, char *msg)
{
    User *u;
    Uid *ud;

    u = finduser(dest);
    ud = find_uid(source);

    send_cmd((ud ? ud->uid : source), "P $%s :%s", (u ? u->uid : dest), msg);
}


void nefarious_cmd_global(char *source, char *buf)
{
    Uid *id;

    if (!buf) {
        return;
    }

    if (source)
      id = find_uid(source);

    send_cmd((source ? (id ? id->uid : p10id) : p10id), "WA :%s", buf);
}

/* GLOBOPS - to handle old WALLOPS */
void nefarious_cmd_global_legacy(char *source, char *fmt)
{
    Uid *id;

    id = find_uid(source);

    send_cmd((id ? id->uid : ServerName), "WA :%s", fmt);
}

int anope_event_sjoin(char *source, int ac, char **av)
{
    do_p10_burst(ac, av);
    return MOD_CONT;
}

int anope_event_account(char *source, int ac, char **av)
{
    NickAlias *na;
    Server *s;
    int res;
    char *cookie;
    char *account;
    char *password;
    char *temp;

    temp = sstrdup(source);

    if (!strcmp(av[1], "C")) {
        cookie = av[ac-3];
        account = ac[av-2];
        password = av[ac-1];

        s = findserver(servlist, temp);
        if (!s)
            s = findserver_uid(servlist, temp);

        if (!(na = findnick(account))) {
             send_cmd(p10id, "AC %s D %s", source, cookie);
             return MOD_CONT;
        }

        res = check_password(password, na->nc->pass);

        if (res == 1) {
            send_cmd(p10id, "AC %s A %s %ld", source, cookie, time(NULL));
        } else {
            send_cmd(p10id, "AC %s D %s", source, cookie);
        }
    }

    return MOD_CONT;
}

int anope_event_nick(char *source, int ac, char **av)
{
    User *user;
    Server *s;
    char *temp;

    temp = sstrdup(source);

    if (ac != 2) {
        char *realname, *ip, *nick;
        char *ident, *host, *modes;
        char *uid = "";
        char *account = "";
        char *fakehost = NULL;
        char *sethost = NULL;
        char *timestamp = "";

        s = findserver(servlist, temp);
        if (!s)
            s = findserver_uid(servlist, temp);
        *source = '\0';

        realname = av[ac-1];
        uid = av[ac-2];
        ip = av[ac-3];
        nick = av[0];
        ident = av[3];
        host = av[4];

        if (strpbrk(av[5], "+")) {
          int cnt = 6;
          int c = 1;
          char *uaccount = "";
          char *accb, *acc = NULL;
          modes = av[5];
          while (*modes) {
              switch (*modes) {
                  case 'r':
                        uaccount = av[cnt++];
                        for (acc = strtok_r(uaccount, ":", &accb);
                             acc;
                             acc = strtok_r(NULL, ":", &accb)) {
                            if (c == 1)
                                account = acc;
                            else if (c == 2)
                                timestamp = acc;
                            c++;
                        }
                        c = 1;
                        break;
                  case 'h':
                    sethost = av[cnt++];
                    break;
                  case 'f':
                    fakehost = av[cnt++];
                    break;
                  default:
                    break;
              }
              modes++;
          }
          modes = av[5];
        } else
          modes = "";

        user = do_nick(source, nick, ident, host, (s ? s->name : temp), realname,
                       strtoul(av[2], NULL, 10), strtoul(timestamp, NULL, 0),
                       base64toIP(ip), NULL, uid, account);

       if (user) {
         if (fakehost || sethost) {
           char *vhost = "";
           if (sethost) {
               int h = 1;
               char *uhb, *uh = NULL;
               char *vident = "";
               for (uh = strtok_r(sethost, "@", &uhb);
                    uh;
                    uh = strtok_r(NULL, "@", &uhb)) {
                   if (h == 1)
                       vident = uh;
                   else if (h == 2)
                       vhost = uh;
                   h++;
               }
               h = 1;
               change_user_username(user, vident);
           }
           change_user_host(user, sethost ? vhost : fakehost);
         }
         nefarious_set_umode(user, 1, &modes);
       }
    } else {
        user = find_byuid(source);
        do_nick(user->nick, av[0], NULL, NULL, NULL, NULL,
                strtoul(av[1], NULL, 10), 0, 0, NULL, NULL, NULL);
    }
    return MOD_CONT;
}

int anope_event_topic(char *source, int ac, char **av)
{
    if (ac == 4) {
        do_topic(source, ac, av);
    } else {
        Channel *c = findchan(av[0]);
        time_t topic_time = time(NULL);

        if (!c) {
            if (debug) {
                alog("debug: TOPIC %s for nonexistent channel %s",
                     merge_args(ac - 1, av + 1), av[0]);
            }
            return MOD_CONT;
        }

        if (check_topiclock(c, topic_time))
            return MOD_CONT;

        if (c->topic) {
            free(c->topic);
            c->topic = NULL;
        }
        if (ac > 1 && *av[ac-1])
            c->topic = sstrdup(av[ac-1]);

        strscpy(c->topic_setter, av[1], sizeof(c->topic_setter));
        c->topic_time = topic_time;

        record_topic(av[0]);
    }
    return MOD_CONT;
}

int anope_event_436(char *source, int ac, char **av)
{
    if (ac < 1)
        return MOD_CONT;

    m_nickcoll(av[0]);
    return MOD_CONT;
}


/* *INDENT-OFF* */
void moduleAddIRCDMsgs(void) {
    Message *m;

    updateProtectDetails("PROTECT","PROTECTME","protect","deprotect","AUTOPROTECT","+","-");

    m = createMessage("ERROR",    anope_event_error); addCoreMessage(IRCD,m);
    m = createMessage("NOTICE",   anope_event_null); addCoreMessage(IRCD,m);
    m = createMessage("SERVER",   anope_event_server); addCoreMessage(IRCD,m);
    m = createMessage("PASS",     anope_event_null); addCoreMessage(IRCD,m);

   /* end of burst */
    m = createMessage("EB",       anope_event_eob); addCoreMessage(IRCD,m);
   /* FAKEHOST */
    m = createMessage("FA",       anope_event_fakehost); addCoreMessage(IRCD,m);
   /* SVSNOOP */
    m = createMessage("SO",       anope_event_null); addCoreMessage(IRCD,m);
    /* NICK */
    m = createMessage("N",        anope_event_nick); addCoreMessage(IRCD,m);
    /* PING */
    m = createMessage("G",        anope_event_ping); addCoreMessage(IRCD,m);
    /* MODE */
    m = createMessage("M",        anope_event_mode); addCoreMessage(IRCD,m);
    /* CREATE */
    m = createMessage("C",        anope_event_create); addCoreMessage(IRCD,m);
    /* JOIN */
    m = createMessage("J",        anope_event_join); addCoreMessage(IRCD,m);
    /* QUIT */
    m = createMessage("Q",        anope_event_quit); addCoreMessage(IRCD,m);
    /* TOPIC */
    m = createMessage("T",        anope_event_topic); addCoreMessage(IRCD,m);
    /* MOTD */
    m = createMessage("MO",       anope_event_motd); addCoreMessage(IRCD,m);
    /* KICK */
    m = createMessage("K",        anope_event_kick); addCoreMessage(IRCD,m);
    /* KILL */
    m = createMessage("D",        anope_event_kill); addCoreMessage(IRCD,m);
    /* AWAY */
    m = createMessage("A",        anope_event_away); addCoreMessage(IRCD,m);
    /* ADMIN */
    m = createMessage("AD",       anope_event_admin); addCoreMessage(IRCD,m);
    /* PART */
    m = createMessage("L",        anope_event_part); addCoreMessage(IRCD,m);
    /* BURST */
    m = createMessage("B",        anope_event_sjoin); addCoreMessage(IRCD,m);
    /* PRIVMSG */
    m = createMessage("P",        anope_event_privmsg); addCoreMessage(IRCD,m);
    /* NOTICE */
    m = createMessage("O",        anope_event_notice); addCoreMessage(IRCD,m);
    /* SQUIT */
    m = createMessage("SQ",       anope_event_squit); addCoreMessage(IRCD,m);
    /* INVITE */
    m = createMessage("I",        anope_event_null); addCoreMessage(IRCD,m);
    /* DESYNCH */
    m = createMessage("DS",       anope_event_error); addCoreMessage(IRCD,m);
    /* WALLOP */
    m = createMessage("WA",       anope_event_null); addCoreMessage(IRCD,m);
    /* VERSION */
    m = createMessage("V",        m_version); addCoreMessage(IRCD,m);
    /* WHOIS */
    m = createMessage("W",        anope_event_whois); addCoreMessage(IRCD,m);
    /* SERVER */
    m = createMessage("S",        anope_event_server); addCoreMessage(IRCD,m);
    /* STATS */
    m = createMessage("R",        m_stats); addCoreMessage(IRCD,m);
    /* ACCCOUNT */
    m = createMessage("AC",       anope_event_account); addCoreMessage(IRCD,m);
    /* GLINE */
    m = createMessage("GL",       anope_event_null); addCoreMessage(IRCD,m);
    /* INFO */
    m = createMessage("F",        anope_event_null); addCoreMessage(IRCD,m);
    /* SETTIME */
    m = createMessage("SE",       anope_event_null); addCoreMessage(IRCD,m);
    /* TIME */
    m = createMessage("TI",       m_time); addCoreMessage(IRCD,m);
    /* TRACE */
    m = createMessage("TR",       anope_event_null); addCoreMessage(IRCD,m);
    /* RPING */
    m = createMessage("RI",       anope_event_null); addCoreMessage(IRCD,m);
    /* End of Burst Acknowledge */
    m = createMessage("EA",       anope_event_null); addCoreMessage(IRCD,m);

    m = createMessage("401",      anope_event_null); addCoreMessage(IRCD,m);
    m = createMessage("402",      anope_event_null); addCoreMessage(IRCD,m);
    m = createMessage("436",      anope_event_436); addCoreMessage(IRCD,m);
}

/* *INDENT-ON* */

int anope_event_fakehost(char *source, int ac, char **av)
{
    User *ud;

    ud = find_byuid(av[0]);

    change_user_host(ud, av[1]);
    return MOD_CONT;
}

void nefarious_cmd_remove_akill(char *user, char *host)
{
    send_cmd(p10id, "GL * -%s", host);
}

void nefarious_cmd_akill(char *user, char *host, char *who, time_t when,
                          time_t expires, char *reason)
{
    Uid *ud;

    ud = find_uid(s_OperServ);

    send_cmd((ud ? ud->uid : s_OperServ), "GL * +%s %ld %ld :%s", host,
             (long int) (expires - (long) time(NULL)), (long int) time(NULL),
             reason);
}


void nefarious_cmd_topic(char *whosets, char *chan, char *whosetit,
                     char *topic, time_t when)
{
    Uid *u;
    u = find_uid(whosets);

    send_cmd((u ? u->uid : whosets), "T %s %s %lu :%s", chan, whosetit,
             (unsigned long int) when, topic);
}

void nefarious_cmd_join(char *user, char *channel, time_t chantime)
{
    Uid *ud;

    ud = find_uid(user);

    if (started) {
         send_cmd((ud ? ud->uid : user), "J %s %ld", channel,
                  chantime);
         send_cmd(p10id, "M %s +o %s %ld", channel, (ud ? ud->uid : user), (long int) time(NULL));
    } else {
        send_cmd(p10id, "B %s %ld %s:o", channel,
                 (long int) time(NULL), (ud ? ud->uid : user));
    }
}

void nefarious_cmd_svskill(char *source, char *user, char *buf)
{
    Uid *ud;
    User *ud2;

    if (!buf) {
        return;
    }

    if (!source || !user) {
        return;
    }

    ud = find_uid(source);
    ud2 = finduser(user);

    send_cmd((ud ? ud->uid : p10id), "D %s :%s", (ud2 ? ud2->uid : user), buf);
}

/* PASS */
void nefarious_cmd_pass(char *pass)
{
    send_cmd(NULL, "PASS :%s", pass);
}

/* SERVER name hop descript */
void nefarious_cmd_server(char *servname, int hop, char *descript)
{
    char newid[3];

    if (p10inuse == 1) {
      int num = atoi(Numeric)+p10servcnt;
      inttobase64(newid, num, 2);
      p10servcnt++;
    }

    send_cmd(((p10inuse == 1) ? p10id : NULL), "%s %s %d %ld %lu J10 %s]]] +s :%s", ((p10inuse == 1) ? "S" : "SERVER"),
               servname, hop,(long int) start_time, (long int) time(NULL), ((p10inuse == 1) ? newid : p10id), descript);

    if (p10inuse == 1) {
      send_cmd(newid, "EB");
      send_cmd(newid, "EA");
    }

    p10inuse = 1;
}
void nefarious_cmd_connect()
{
    char *hops = NULL;

    if (!BadPtr(Numeric)) {
        hops = sstrdup("0");
        inttobase64(p10id, atoi(Numeric), 2);
        me_server = new_server(NULL, ServerName, ServerDesc, SERVER_ISME, p10id);
        nefarious_cmd_pass(RemotePassword);
        nefarious_cmd_server(ServerName, 1, ServerDesc);
    } else {
        alog("debug: Numeric is required for P10 ircds");
        exit(1);
    }
}

void nefarious_cmd_bot_nick(char *nick, char *user, char *host, char *real,
                         char *modes)
{
    char nicknumbuf[6];
    send_cmd(p10id, "N %s 1 %ld %s %s %s B]AAAB %sAA%c :%s", nick,
             (long int) time(NULL), user, host, modes, p10id, (p10nickcnt + 'A'),
             real);
    snprintf(nicknumbuf, 6, "%sAA%c", p10id, (p10nickcnt + 'A'));
    new_uid(nick, nicknumbuf);
    p10nickcnt++;

}

void nefarious_cmd_part(char *nick, char *chan, char *buf)
{
    Uid *u;

    u = find_uid(nick);

    if (buf) {
        send_cmd((u ? u->uid : nick), "L %s :%s", chan, buf);
    } else {
        send_cmd((u ? u->uid : nick), "L %s", chan);
    }
}

int anope_event_ping(char *source, int ac, char **av)
{

    if (ac < 1)
        return MOD_CONT;
    nefarious_cmd_pong(p10id, av[0]);
    return MOD_CONT;
}

int anope_event_away(char *source, int ac, char **av)
{
    User *u;

    if (!source) {
        return MOD_CONT;
    }

    u = find_byuid(source);

    if (!ac) {
       m_away((u ? u->nick : source), NULL);
    } else {
       m_away((u ? u->nick : source), av[0]);
    }
    return MOD_CONT;
}

int anope_event_kill(char *source, int ac, char **av)
{
    if (ac != 2)
        return MOD_CONT;

    m_kill(av[0], av[1]);
    return MOD_CONT;
}

int anope_event_kick(char *source, int ac, char **av)
{
    if (ac != 3)
        return MOD_CONT;
    do_p10_kick(source, ac, av);
    return MOD_CONT;
}

int anope_event_eob(char *source, int ac, char **av)
{
    Server *s;
    s = findserver(servlist, source);
    /* If we found a server with the given source, that one just
     * finished bursting. If there was no source, then our uplink
     * server finished bursting. -GD
     */
    if (!s && serv_uplink)
        s = serv_uplink;
    finish_sync(s, 1);

    send_cmd(p10id, "EA");
    return MOD_CONT;
}

void nefarious_cmd_eob()
{
    send_cmd(p10id, "EB");
}


int anope_event_join(char *source, int ac, char **av)
{
    User *u;

    u = find_byuid(source);

    if (ac != 2) {
        return MOD_CONT;
    }

    do_join((u ? u->nick : source), ac, av);
    return MOD_CONT;
}

int anope_event_create(char *source, int ac, char **av)
{
    char nick[BUFSIZE] = "";
    char *cav[3];
    User *u;

    u = find_byuid(source);

    cav[0] = av[1];
    cav[1] = av[0];
    strcat(nick, "@");
    strcat(nick, u->nick);
    cav[2] = nick;

    do_sjoin((u ? u->nick : source), 3, cav);

    return MOD_CONT;
}

int anope_event_motd(char *source, int ac, char **av)
{
    if (!source) {
        return MOD_CONT;
    }

    m_motd(source);
    return MOD_CONT;
}

int anope_event_privmsg(char *source, int ac, char **av)
{
    User *u;
    Uid *id;

    u = find_byuid(source);
    id = find_nickuid(av[0]);

    if (ac != 2)
        return MOD_CONT;
    m_privmsg((u ? u->nick : source), (id ? id->nick : av[0]), av[1]);
    return MOD_CONT;
}

int anope_event_part(char *source, int ac, char **av)
{
    User *u;

    u = find_byuid(source);

    if (ac < 1 || ac > 2)
        return MOD_CONT;
    do_part((u ? u->nick : source), ac, av);
    return MOD_CONT;
}

int anope_event_whois(char *source, int ac, char **av)
{
    if (source && ac >= 1) {
        m_whois(source, av[1]);
    }
    return MOD_CONT;
}

/* EVENT: SERVER */
int anope_event_server(char *source, int ac, char **av)
{
    Server *s;
    char uplinknum[5] = "";
    *uplinknum = '\0';
    strcpy(uplinknum, av[5]);
    uplinknum[2] = '\0';

    if (!stricmp(av[1], "1")) {
        uplink = sstrdup(av[0]);
        do_server(source, av[0], av[1], av[7], uplinknum);
    } else {
        s = findserver_uid(servlist, source);
        do_server((s ? s->name : source), av[0], av[1], av[7], uplinknum);
    }
    return MOD_CONT;
}

int anope_event_squit(char *source, int ac, char **av)
{
    if (ac != 3)
        return MOD_CONT;
    do_squit(source, ac, av);
    return MOD_CONT;
}

int anope_event_quit(char *source, int ac, char **av)
{
    if (ac != 1)
        return MOD_CONT;
    do_quit(source, ac, av);
    return MOD_CONT;
}

void nefarious_cmd_372(char *source, char *msg)
{
    send_cmd(p10id, "372 %s :- %s", source, msg);
}

void nefarious_cmd_372_error(char *source)
{
    send_cmd(p10id, "422 %s :- MOTD file not found!  Please "
             "contact your IRC administrator.", source);
}

void nefarious_cmd_375(char *source)
{
    send_cmd(p10id, "375 %s :- %s Message of the Day",
             source, ServerName);
}

void nefarious_cmd_376(char *source)
{
    send_cmd(p10id, "376 %s :End of /MOTD command.", source);
}

/* 391 */
void nefarious_cmd_391(char *source, char *timestr)
{
    if (!timestr) {
        return;
    }
    send_cmd(p10id, "391 %s %s :%s", source, ServerName, timestr);
}

/* 250 */
void nefarious_cmd_250(char *buf)
{
    if (!buf) {
        return;
    }

    send_cmd(p10id, "250 %s", buf);
}

/* 307 */
void nefarious_cmd_307(char *buf)
{
    if (!buf) {
        return;
    }

    send_cmd(p10id, "307 %s", buf);
}

/* 311 */
void nefarious_cmd_311(char *buf)
{
    if (!buf) {
        return;
    }

    send_cmd(p10id, "311 %s", buf);
}

/* 312 */
void nefarious_cmd_312(char *buf)
{
    if (!buf) {
        return;
    }

    send_cmd(p10id, "312 %s", buf);
}

/* 317 */
void nefarious_cmd_317(char *buf)
{
    if (!buf) {
        return;
    }

    send_cmd(p10id, "317 %s", buf);
}

/* 219 */
void nefarious_cmd_219(char *source, char *letter)
{
    if (!source) {
        return;
    }

    if (letter) {
        send_cmd(p10id, "219 %s %c :End of /STATS report.", source,
                 *letter);
    } else {
        send_cmd(p10id, "219 %s l :End of /STATS report.", source);
    }
}

/* 401 */
void nefarious_cmd_401(char *source, char *who)
{
    if (!source || !who) {
        return;
    }
    send_cmd(p10id, "401 %s %s :No such service.", source, who);
}

/* 318 */
void nefarious_cmd_318(char *source, char *who)
{
    if (!source || !who) {
        return;
    }

    send_cmd(p10id, "318 %s %s :End of /WHOIS list.", source, who);
}

/* 242 */
void nefarious_cmd_242(char *buf)
{
    if (!buf) {
        return;
    }

    send_cmd(p10id, "242 %s", buf);
}

/* 243 */
void nefarious_cmd_243(char *buf)
{
    if (!buf) {
        return;
    }

    send_cmd(p10id, "243 %s", buf);
}

/* 211 */
void nefarious_cmd_211(char *buf)
{
    if (!buf) {
        return;
    }

    send_cmd(p10id, "211 %s", buf);
}

void nefarious_cmd_mode(char *source, char *dest, char *buf)
{
    Uid *u;

    u = find_uid(source);

    if (!buf) {
        return;
    }
    send_cmd((u ? u->uid : (source ? source : p10id)), "M %s %s %ld", dest, buf, (long int) time(NULL));
}

void nefarious_cmd_nick(char *nick, char *name, char *mode)
{
    char nicknumbuf[6];

    send_cmd(p10id, "N %s 1 %ld %s %s %s B]AAAB %sAA%c :%s", nick,
             (long int) time(NULL), ServiceUser, ServiceHost, mode, p10id,
             (p10nickcnt + 'A'), name);
    snprintf(nicknumbuf, 6, "%sAA%c", p10id, (p10nickcnt + 'A'));
    new_uid(nick, nicknumbuf);
    p10nickcnt++;
}

void nefarious_cmd_kick(char *source, char *chan, char *user, char *buf)
{
    Uid *ud;
    User *tgt;
    ud = find_uid(source);
    tgt = finduser(user);

    if (buf) {
        send_cmd((ud ? ud->uid : source), "K %s %s :%s", chan, (tgt ? tgt->uid : user), buf);
    } else {
        send_cmd((ud ? ud->uid : source), "K %s %s", chan, (tgt ? tgt->uid : user));
    }
}

void nefarious_cmd_notice_ops(char *source, char *dest, char *buf)
{
    Uid *u;

    u = find_uid(source);

    if (!buf) {
        return;
    }

    send_cmd((u ? u->uid : source), "O @%s :%s", dest, buf);
}

void nefarious_cmd_bot_chan_mode(char *nick, char *chan)
{
    anope_cmd_mode(nick, chan, "%s %s", ircd->botchanumode, nick);
}

/* QUIT */
void nefarious_cmd_quit(char *source, char *buf)
{
    Uid *ud;
    ud = find_uid(source);

    if (buf) {
        send_cmd((ud ? ud->uid : source), "QUIT :%s", buf);
    } else {
        send_cmd((ud ? ud->uid : source), "QUIT");
    }
}

/* PONG */
void nefarious_cmd_pong(char *servname, char *who)
{
    char *t, *s;
    uint32 ts, tsnow, value;
    Server *sv;

    if (!servname || !who) {
        return;
    }

    sv = findserver(servlist, servname);
    if (!sv)
      sv = findserver_uid(servlist, servname);

    t = myStrGetToken(who, '!', 1);
    tsnow = time(NULL);
    if (t) {

        s = myStrGetToken(t, '.', 0);
        ts = strtol(s, NULL, 10);
        value = tsnow - ts; 

        send_cmd((sv ? sv->suid : p10id), "Z %s %ld %ld %ld %s", (sv ? sv->suid : p10id), (long int) ts, (long int) tsnow, (long int) value, militime_float(NULL));
    } else /* cmd_pong from functions like the mysql bits */
        send_cmd((sv ? sv->suid : p10id), "Z %s %ld 0 %ld %s", (sv ? sv->suid : p10id), (long int) tsnow, (long int) tsnow, militime_float(NULL));
}

/* INVITE */
void nefarious_cmd_invite(char *source, char *chan, char *nick)
{
    Uid *ud;
    ud = find_uid(source);

    if (!source || !chan || !nick) {
        return;
    }

    send_cmd((ud ? ud->uid : source), "I %s %s", nick, chan);
}

/* SQUIT */
void nefarious_cmd_squit(char *servname, char *message)
{
    Server *s;

    if (!servname || !message) {
        return;
    }
    s = findserver(servlist, servname);
    if (!s)
      s = findserver_uid(servlist, servname);

    if (!s)
      return;

    send_cmd(p10id, "SQ %s :%s", s->suid, message);
}

int anope_event_mode(char *source, int ac, char **av)
{
    User *u;
    char *sender;
    u = find_byuid(source);

    if (!u) {
        sender = source;
    } else {
        sender = u->nick;
    }

    if (ac < 2)
        return MOD_CONT;

    if (*av[0] == '#' || *av[0] == '&') {
        do_cmode(source, ac, av);
    } else {
        do_umode(sender, ac, av);
    }
    return MOD_CONT;
}

void nefarious_cmd_351(char *source)
{
    send_cmd(p10id, "351 %s Anope-%s %s :%s - %s -- %s",
             source, version_number, ServerName, ircd->name, version_flags,
             version_build);
}

void nefarious_cmd_nc_change(User * u)
{
    /* not supported */
}

void nefarious_cmd_svid_umode(char *nick, time_t ts)
{
    /* not supported */
}

void nefarious_cmd_svid_umode2(User * u, char *ts)
{
    /* not supported */
}

void nefarious_cmd_svid_umode3(User * u, char *ts)
{
    /* not supported */
}

void nefarious_cmd_svsnick(char *nick, char *newnick, time_t when)
{
    User *oldnick;

    if (!nick || !newnick) {
        return;
    }

    oldnick = finduser(nick);

    send_cmd(p10id, "SN %s %s", (oldnick ? oldnick->uid :nick), newnick);
}

void nefarious_cmd_guest_nick(char *nick, char *user, char *host, char *real,
                           char *modes)
{
    char nicknumbuf[6];
    send_cmd(p10id, "N %s 1 %ld %s %s %s B]AAAB %sAA%c :%s", nick,
             (long int) time(NULL), user, host, modes, p10id,
             (p10nickcnt + 'A'), real);
    snprintf(nicknumbuf, 6, "%sAA%c", p10id, (p10nickcnt + 'A'));
    new_uid(nick, nicknumbuf);
    p10nickcnt++;
}

void nefarious_cmd_svsmode(User * u, int ac, char **av)
{
    if (ac >= 1) {
        if (!u || !av[0]) {
            return;
        }
        send_cmd(ServerName, "M %s %s%s%s", u->nick, av[0],
                 (ac == 2 ? " " : ""), (ac == 2 ? av[1] : ""));
    }
}

void nefarious_cmd_chg_nick(char *oldnick, char *newnick)
{
    Uid *ud;

    if (!oldnick || !newnick) {
        return;
    }

    ud = find_uid(oldnick);

    send_cmd((ud ? ud->uid : oldnick), "N %s", newnick);
}

int anope_event_pass(char *source, int ac, char **av)
{
    /* currently not used but removes the message : unknown message from server */
    return MOD_CONT;
}

void nefarious_cmd_svsjoin(char *source, char *nick, char *chan)
{
    User *tnick;

    tnick = finduser(nick);

    if (!tnick) {
        return;
    }

    send_cmd(p10id, "SJ %s :%s", (tnick ? tnick->uid : nick), chan);
}

void nefarious_cmd_svspart(char *source, char *nick, char *chan)
{
    User *tnick;

    tnick = finduser(nick);

    if (!tnick) {
        return;
    }

    send_cmd(p10id, "SP %s :%s", (tnick ? tnick->uid : nick), chan);
}

void nefarious_cmd_swhois(char *source, char *who, char *mask)
{
    User *tnick;

    tnick = finduser(who);

    if (!tnick) {
        return;
    }

    send_cmd(p10id, "SW %s :%s", (tnick ? tnick->uid : who), mask);
}

int anope_event_notice(char *source, int ac, char **av)
{
    /* not used */
    return MOD_CONT;
}

int anope_event_admin(char *source, int ac, char **av)
{
    /* not used */
    return MOD_CONT;
}

int nefarious_flood_mode_check(char *value)
{
    /* not used */
    return 0;
}

int anope_event_error(char *source, int ac, char **av)
{
    if (ac >= 1) {
        if (debug) {
            alog("debug: %s", av[0]);
        }
    }
    return MOD_CONT;
}

void nefarious_cmd_jupe(char *jserver, char *who, char *reason)
{
    char rbuf[256];

    snprintf(rbuf, sizeof(rbuf), "Juped by %s%s%s", who,
             reason ? ": " : "", reason ? reason : "");

    nefarious_cmd_squit(jserver, rbuf);
    nefarious_cmd_server(jserver, 2, rbuf);
    new_server(me_server, jserver, rbuf, SERVER_JUPED, NULL);
}

/* 
  1 = valid nick
  0 = nick is in valid
*/
int nefarious_valid_nick(char *nick)
{
    /* no hard coded invalid nicks */
    return 1;
}

void nefarious_cmd_account(User * u)
{
    send_cmd(p10id, "AC %s R %s %ld", u->uid, u->nick, (long int) time(NULL));
}

void nefarious_cmd_ctcp(char *source, char *dest, char *buf)
{
    char *s;
    Uid *u;

    u = find_uid(source);

    if (!buf) {
        return;
    } else {
        s = normalizeBuffer(buf);
    }

    send_cmd((u ? u->uid : source), "O %s :\1%s \1", dest, s);
    free(s);
}

void nefarious_cmd_svsnoop(char *server, int set)
{
    Server *s;

    s = findserver(servlist, server);
    if (s)
      send_cmd(p10id, "SO %s %s", s->suid, (set ? "+" : "-"));
}

void nefarious_cmd_unsqline(char *user)
{
    /* not supported */
}

void nefarious_cmd_sqline(char *mask, char *reason)
{
    /* not supported */
}

void nefarious_cmd_vhost_off(User * u)
{
    send_cmd(p10id, "M %s -h", (u->uid ? u->uid : u->nick));
}

void nefarious_cmd_vhost_on(char *nick, char *vIdent, char *vhost)
{
    User *u;
    u = finduser(nick);

    if (vIdent)
      send_cmd(p10id, "M %s +h %s@%s", (u ? u->uid : nick), vIdent, vhost);
    else
      send_cmd(p10id, "M %s +h %s@%s", (u ? u->uid : nick), u->username, vhost);

}

/**
 * Tell anope which function we want to perform each task inside of anope.
 * These prototypes must match what anope expects.
 **/
void moduleAddAnopeCmds()
{
    pmodule_cmd_remove_akill(nefarious_cmd_remove_akill);
    pmodule_cmd_akill(nefarious_cmd_akill);
    pmodule_cmd_vhost_off(nefarious_cmd_vhost_off);
    pmodule_cmd_vhost_on(nefarious_cmd_vhost_on);
    pmodule_cmd_svsmode(nefarious_cmd_svsmode);
    pmodule_cmd_svsnoop(nefarious_cmd_svsnoop);
    pmodule_cmd_topic(nefarious_cmd_topic);
    pmodule_cmd_svskill(nefarious_cmd_svskill);
    pmodule_cmd_372(nefarious_cmd_372);
    pmodule_cmd_372_error(nefarious_cmd_372_error);
    pmodule_cmd_375(nefarious_cmd_375);
    pmodule_cmd_376(nefarious_cmd_376);
    pmodule_cmd_nick(nefarious_cmd_nick);
    pmodule_cmd_guest_nick(nefarious_cmd_guest_nick);
    pmodule_cmd_mode(nefarious_cmd_mode);
    pmodule_cmd_bot_nick(nefarious_cmd_bot_nick);
    pmodule_cmd_kick(nefarious_cmd_kick);
    pmodule_cmd_notice_ops(nefarious_cmd_notice_ops);
    pmodule_cmd_notice(nefarious_cmd_notice);
    pmodule_cmd_notice2(nefarious_cmd_notice2);
    pmodule_cmd_privmsg(nefarious_cmd_privmsg);
    pmodule_cmd_privmsg2(nefarious_cmd_privmsg2);
    pmodule_cmd_serv_notice(nefarious_cmd_serv_notice);
    pmodule_cmd_serv_privmsg(nefarious_cmd_serv_privmsg);
    pmodule_cmd_bot_chan_mode(nefarious_cmd_bot_chan_mode);
    pmodule_cmd_351(nefarious_cmd_351);
    pmodule_cmd_quit(nefarious_cmd_quit);
    pmodule_cmd_pong(nefarious_cmd_pong);
    pmodule_cmd_join(nefarious_cmd_join);
    pmodule_cmd_invite(nefarious_cmd_invite);
    pmodule_cmd_part(nefarious_cmd_part);
    pmodule_cmd_391(nefarious_cmd_391);
    pmodule_cmd_250(nefarious_cmd_250);
    pmodule_cmd_307(nefarious_cmd_307);
    pmodule_cmd_311(nefarious_cmd_311);
    pmodule_cmd_312(nefarious_cmd_312);
    pmodule_cmd_317(nefarious_cmd_317);
    pmodule_cmd_219(nefarious_cmd_219);
    pmodule_cmd_401(nefarious_cmd_401);
    pmodule_cmd_318(nefarious_cmd_318);
    pmodule_cmd_242(nefarious_cmd_242);
    pmodule_cmd_243(nefarious_cmd_243);
    pmodule_cmd_211(nefarious_cmd_211);
    pmodule_cmd_global(nefarious_cmd_global);
    pmodule_cmd_global_legacy(nefarious_cmd_global_legacy);
    pmodule_cmd_sqline(nefarious_cmd_sqline);
    pmodule_cmd_squit(nefarious_cmd_squit);
    pmodule_cmd_chg_nick(nefarious_cmd_chg_nick);
    pmodule_cmd_svsnick(nefarious_cmd_svsnick);
    pmodule_cmd_connect(nefarious_cmd_connect);
    pmodule_cmd_nc_change(nefarious_cmd_nc_change);
    pmodule_cmd_svid_umode(nefarious_cmd_svid_umode);
    pmodule_cmd_svid_umode2(nefarious_cmd_svid_umode2);
    pmodule_cmd_svid_umode3(nefarious_cmd_svid_umode3);
    pmodule_cmd_eob(nefarious_cmd_eob);
    pmodule_flood_mode_check(nefarious_flood_mode_check);
    pmodule_cmd_jupe(nefarious_cmd_jupe);
    pmodule_valid_nick(nefarious_valid_nick);
    pmodule_cmd_ctcp(nefarious_cmd_ctcp);
    pmodule_cmd_account(nefarious_cmd_account);
    pmodule_set_umode(nefarious_set_umode);
}

/** 
 * Now tell anope how to use us.
 **/
int AnopeInit(int argc, char **argv)
{

    moduleAddAuthor("Neil Spierling <sirvulcan@gmail.com>");
    moduleAddVersion("$Id: nefarious.c 1393 2005-12-05 04:39:41Z sirvulcan $");
    moduleSetType(PROTOCOL);

    pmodule_ircd_version("Nefarious IRCu 0.5.0+");
    pmodule_ircd_cap(myIrcdcap);
    pmodule_ircd_var(myIrcd);
    pmodule_ircd_cbmodeinfos(myCbmodeinfos);
    pmodule_ircd_cumodes(myCumodes);
    pmodule_ircd_flood_mode_char_set("");
    pmodule_ircd_flood_mode_char_remove("");
    pmodule_ircd_cbmodes(myCbmodes);
    pmodule_ircd_cmmodes(myCmmodes);
    pmodule_ircd_csmodes(myCsmodes);
    pmodule_ircd_useTSMode(0);

        /** Deal with modes anope _needs_ to know **/
    pmodule_invis_umode(UMODE_i);
    pmodule_oper_umode(UMODE_o);
    pmodule_invite_cmode(CMODE_i);
    pmodule_secret_cmode(CMODE_s);
    pmodule_private_cmode(CMODE_p);
    pmodule_key_mode(CMODE_k);
    pmodule_limit_mode(CMODE_l);

    moduleAddAnopeCmds();
    moduleAddIRCDMsgs();

    return MOD_CONT;
}
